#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"A Proxomitron Helper Program"

__author__ = 'phoenix'
__version__ = '0.5'

from http.server import HTTPServer, BaseHTTPRequestHandler
from socketserver import ThreadingMixIn
# Not using urllib becaust it follows redirection
from http.client import HTTPConnection, HTTPSConnection
from urllib.parse import urlparse

import configparser, threading, ssl
from proxytool import ProxyRequestHandler

class LoadConfig:
    def __init__(self, configfile):
        self.config = configparser.ConfigParser()
        self.config.read(configfile)
        self.PROXADDR = self.config['GENERAL'].get('ProxAddr')
        self.FRONTPORT = int(self.config['GENERAL'].get('FrontPort'))
        self.REARPORT = int(self.config['GENERAL'].get('RearPort'))
        self.HTTPSPROXY = self.config['GENERAL'].get('HTTPSProxy')

class FrontServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""
    pass

class RearServer(ThreadingMixIn, HTTPServer):
    """Handle requests in a separate thread."""
    pass

class FrontRequestHandler(ProxyRequestHandler):
    """
    Sit between the client and Proxomitron
    Convert https request to http
    """
    server_version = "Proxomitron FrontProxy/%s" % __version__

    def do_METHOD(self):
        "Forward request to Proxomitron"
        host = self.host if self.port == '443' else "%s:%s" % (self.host, self.port)
        path = "http://%s%s" % (host, self.path)
        data_length = self.headers.get("Content-Length")
        data = self.rfile.read(int(data_length)) if data_length else None
        # signal Server to close the connection
        self.headers.__delitem__("Connection")
        self.headers["Connection"] = "close"        
        # Tag the request so Proxomitron can recognize it
        self.headers["Tagged"] = self.version_string()
        conn = HTTPConnection(config.PROXADDR)
        conn.request(self.command, path, data, self.headers)
        response = conn.getresponse()
        self.send_response_only(response.status, response.reason)
        self.flush_headers()
        self.wfile.write(response.msg.as_bytes())
        self.wfile.write(response.read())

    do_GET = do_POST = do_HEAD = do_PUT = do_DELETE = do_METHOD

class RearRequestHandler(ProxyRequestHandler):
    """
    Supposed to be the parent proxy for Proxomitron for tagged requests
    Convert http request to https
    
    """
    server_version = "Proxomitron RearProxy/%s" % __version__
    
    def do_METHOD(self):
        "Convert http request to https"
        # request line: GET http://somehost.com/path?attr=value HTTP/1.1
        parsed_url = urlparse(self.path)
        path = "/" + self.path[7:].split('/', 1)[1]
        data_length = self.headers.get("Content-Length")
        data = self.rfile.read(int(data_length)) if data_length else None
        # Remove the tag
        del self.headers["Tagged"]
        conn = HTTPSConnection(parsed_url.hostname, parsed_url.port, context=ssl_context)
        if config.HTTPSPROXY:
            conn = HTTPSConnection(config.HTTPSPROXY, context=ssl_context)
            conn.set_tunnel(parsed_url.hostname, parsed_url.port)
        conn.request(self.command, path, data, self.headers)
        response = conn.getresponse()
        self.log_request(response.status)
        self.send_response_only(response.status, response.reason)
        self.flush_headers()
        # content is to be pushed NOT chunked
        response.msg.__delitem__("Transfer-Encoding")
        # signal Client to close the connection
        response.msg.__delitem__("Connection")
        response.msg["Connection"] = "close"
        self.wfile.write(response.msg.as_bytes())
        self.wfile.write(response.read())

    do_GET = do_POST = do_HEAD = do_PUT = do_DELETE = do_METHOD
    
try:
    config = LoadConfig('ProxHTTPSProxy.ini')
    ssl_context = ssl.create_default_context()
    
    frontserver = FrontServer(('', config.FRONTPORT), FrontRequestHandler)
    rearserver = RearServer(('', config.REARPORT), RearRequestHandler)
    for server in (frontserver, rearserver):
          thread = threading.Thread(target=server.serve_forever)
          thread.dameon = True
          thread.start()

    print("=" * 76)
    print('ProxHTTPSProxy %s serving now, <Ctrl-C> to stop ...' % __version__)
    print()
    print('  FrontServer  : localhost:%s' % config.FRONTPORT)
    print('  RearServer   : localhost:%s' % config.REARPORT)
    print('  ParentServer : %s' % config.HTTPSPROXY)
    print('  Proxomitron  : ' + config.PROXADDR)
    print("=" * 76)
except KeyboardInterrupt:
    print("Quitting...")
